/*
 * Copyright (C)  1999  Warren Jasper
 * All rights reserved.
 *
 *   Documentation:
 *       PCI-DIO24 Manual Revision 1, September 1998, ComputerBoards Inc.
 *       125 High Street, #6, Mansfield, MA 02048
 *       (508) 261-1123 www.computerboards.com
 *       
 *       PLX Technology
 *       www.plxtech.com  PCI-9050-1 Data Boot
 *
 *       Intel: 8255A-5 Programmable Peripheral Interface
 *       Order Number 231308-004
 *       www.intel.com
 *
 */

/***************************************************************************
 *
 * pci-dio24.c
 *
 ***************************************************************************/

#define MODULE

/* retrieve the CONFIG_* macros */
#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/timer.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/signal.h>
#include <linux/config.h>       /* for CONFIG_PCI */
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include "dio24.h"
#include "dio.h"

#ifdef CONFIG_PCI
#include <linux/pci.h>
#include <linux/bios32.h>
#endif

/***************************************************************************
 *
 * Prototype of public and private functions.
 *
 ***************************************************************************/

static int dio_read(struct inode *iNode, struct file *filePtr, char *buf, int count);
static int dio_write(struct inode *iNode, struct file *filePtr, const char *buf, int count);
static int dio_open(struct inode *iNode, struct file *filePtr);
static void dio_close(struct inode *iNode, struct file *filePtr);
static int dio_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg);
static int config_region( unsigned int pci_ioaddr, int size, WORD *base );
static void dio24_ReadInterrupt();

int init_module(void);
void cleanup_module(void) ;

/***************************************************************************
 *
 * Global data.
 *
 ***************************************************************************/

static WORD base1;                          /* base register 1 address       */
static WORD base2;                          /* base register 2 address       */
static int MajorNumber = DEFAULT_MAJOR_DEV; /* Major number compiled in      */
static BoardRec BoardData;                  /* Board specific information    */
static ChanRec Chan[DIO_PORTS];             /* Channel specific information  */
static BYTE dio_reg = 0x80;                 /* status register 1             */
struct wait_queue  *dio24_wait;             /* wait semaphore                */


/***************************************************************************
 *
 *
 ***************************************************************************/

static struct file_operations dio_fops = {
	NULL,           /* dio_seek */
	dio_read,       /* dio_read */
	dio_write,      /* dio_write */
	NULL,		/* dio_readdir */
	NULL,		/* dio_select */
	dio_ioctl,	/* dio_ioctl */
	NULL,
	dio_open,
	dio_close,
	NULL		/* fsync */
};

#ifdef __cplusplus
extern "C" {
#endif

 /*********************************************************************************
 *                                                                                *
 * Loads driver. Called when "insmod pci-dio24.o" is invoked on the command line. *
 *               The board is set to irq if enabled                               *
 *                                                                                *
 **********************************************************************************/

int init_module(void) 
{
  register int i;
  int pci_index;
  unsigned char pci_bus, pci_device_fn;
  unsigned int pci_ioaddr;
  unsigned char pci_irq_line;
  int err;

#ifndef CONFIG_PCI
    /* this should not happen. */
    printk("%s pci not configured with kernel!\n", ADAPTER_ID);
    return -ENODEV;
#endif

  /* Register as a device with kernel.  */

  if ((err = register_chrdev(MajorNumber, "pci-dio24", &dio_fops))) {
      printk("%s: Failure to load module. error %d\n", ADAPTER_ID, -err);
      return err;
  } 

  /* Register the region of iospace with kernel and get base address */

  if(pcibios_present()) {
    for ( pci_index = 0; pci_index < MAX_BOARDS; pci_index++ ) {
      if ( pcibios_find_device (PCI_VENDOR_ID_CBOARDS,
              PCI_DEVICE_ID_CBOARDS_DIO24, pci_index, &pci_bus, 
              &pci_device_fn) != 0 ) break;

      /* Get PCI_BASE_ADDRESS_1 */
      pcibios_read_config_dword( pci_bus, pci_device_fn,
                            PCI_BASE_ADDRESS_1, &pci_ioaddr );
      /* Strip the I/O address out of the returned value */
      pci_ioaddr &= PCI_BASE_ADDRESS_IO_MASK;
      if (config_region( pci_ioaddr, BADR1_SIZE, &base1 ) != 0 ) {
        printk("%s: Can't allocate region port address %#x\n", 
                ADAPTER_ID, pci_ioaddr );
        if (unregister_chrdev(MajorNumber, "pci-dio24") != 0) {
          printk("%s: unregister_chrdev() failed.\n", ADAPTER_ID);
        }
        return -ENODEV;
      }
      if ( pci_ioaddr != base1 ) {
          pci_ioaddr = base1 | 0x1;
          pcibios_write_config_dword( pci_bus, pci_device_fn,
                        PCI_BASE_ADDRESS_1, pci_ioaddr );
      }

      /* Get PCI_BASE_ADDRESS_2 */
      pcibios_read_config_dword( pci_bus, pci_device_fn,
                                 PCI_BASE_ADDRESS_2, &pci_ioaddr );
      /* Strip the I/O address out of the returned value */
      pci_ioaddr &= PCI_BASE_ADDRESS_IO_MASK;
      if (config_region( pci_ioaddr, BADR2_SIZE, &base2 ) != 0 ) {
        printk("%s: Can't allocate region port address %#x\n", 
                ADAPTER_ID, pci_ioaddr );
        if (unregister_chrdev(MajorNumber, "pci-dio24") != 0) {
          printk("%s: unregister_chrdev() failed.\n", ADAPTER_ID);
        }
        return -ENODEV;
      }
      if ( pci_ioaddr != base2 ) {
          pci_ioaddr = base2 | 0x1;
          pcibios_write_config_dword( pci_bus, pci_device_fn,
                       PCI_BASE_ADDRESS_2, pci_ioaddr );
      }

      if ( ENABLE_INTERRUPTS ) {
        /* Register interrupt handler. */

        pcibios_read_config_byte( pci_bus, pci_device_fn, PCI_INTERRUPT_LINE, &pci_irq_line );
        BoardData.irq = pci_irq_line;
        if (request_irq(pci_irq_line, dio24_ReadInterrupt, (SA_INTERRUPT | SA_SHIRQ), 
            "pci-dio24", (void *) PCI_VENDOR_ID_CBOARDS)) {
          /* No free irq found! cleanup and exit */
          printk("%s: Can't request IRQ %d\n", ADAPTER_ID, BoardData.irq);
          if (unregister_chrdev(MajorNumber, "pci-dio24") != 0) { 
            printk("%s: unregister_chrdev() failed.\n", ADAPTER_ID);
          }
          release_region( base1,  BADR1_SIZE );       /* release region */
          release_region( base2,  BADR2_SIZE );       /* release region */
          return -EADDRNOTAVAIL;
          break;
        }
      }
    }

    if ( pci_index == 0 ) {
      printk("%s No device present.\n", ADAPTER_ID);
      return -ENODEV;
    }
  } else {
    printk("%s No pci bios present.\n", ADAPTER_ID);
    return -ENODEV;
  }

  BoardData.base1 = base1;
  BoardData.base2 = base2;

  BoardData.control_reg = inl(PCI_INT_ADDR);

 
  if ( ENABLE_INTERRUPTS ) {
    BoardData.control_reg |= (INTE | PCIINT);
    outl(BoardData.control_reg, PCI_INT_ADDR);
    printk("%s: BADR1=%#x  BADR2=%#x  IRQ=%d.",ADAPTER_ID, base1, base2,  BoardData.irq );
  } else {
    BoardData.control_reg &= ~(INTE | PCIINT);
    outl(BoardData.control_reg, PCI_INT_ADDR);
    printk("%s: BADR1=%#x  BADR2=%#x.", ADAPTER_ID, base1, base2);
  }
  printk(" 9/9/99 wjasper@tx.ncsu.edu\n");

  /* Set all channel structures to show nothing active/open */

  Chan[0].addr = DIO_PORTA;
  Chan[1].addr = DIO_PORTB;
  Chan[2].addr = DIO_PORTC;

  for (i = 0; i < DIO_PORTS; i++) {
      Chan[i].open = FALSE;
      Chan[i].mode = 0;
      outb_p(0x0, Chan[i].addr);
  }

  return 0;
}

/***************************************************************************
 *
 * Test memory locations for exsistance of pci-dio24 board.
 *
 ***************************************************************************/
static int config_region( unsigned int pci_ioaddr, int size, WORD *base )
{

  /* check to see if the allocated region is ok */
  if ( check_region(pci_ioaddr, size) == 0 ) {
    request_region( pci_ioaddr, size, "pci-dio24" );
    *base = pci_ioaddr;
    return 0;
  } else {
    /* Look to see if there is another place in the io space */
    for ( pci_ioaddr = PORT_MIN; pci_ioaddr < PORT_MAX; pci_ioaddr += PORT_STEP ) {
      if ( check_region(pci_ioaddr,  size ) == 0 ) {
        request_region( pci_ioaddr, size, "pci-dio24" );
        *base = pci_ioaddr;
        return 0;
      }
    }
    
    if ( pci_ioaddr > PORT_MAX ) {
      printk("%s: No free address from %#x to %#x.\n", 
             ADAPTER_ID, PORT_MIN, PORT_MAX);
      if (unregister_chrdev(MajorNumber, "pci-dio24") != 0) {
          printk("%s: unregister_chrdev() failed.\n", ADAPTER_ID);
      }
      return -ENODEV;
    }
  }
  return 1;
}


/***************************************************************************
 *
 * Remove driver. Called when "rmmod pci-dio24" is run on the command line.
 *
 ***************************************************************************/

void cleanup_module(void) 
{
  if (MOD_IN_USE) {
    printk("%s: device busy, remove delayed\n", ADAPTER_ID);
    return;
  }

  release_region( base1,  BADR1_SIZE );       /* release region */
  release_region( base2,  BADR2_SIZE );       /* release region */

  if ( ENABLE_INTERRUPTS ) {
    free_irq(BoardData.irq, NULL);
  }

  if (unregister_chrdev(MajorNumber, "pci-dio24") != 0) {
    printk("%s: cleanup_module failed\n", ADAPTER_ID);
  } else {

  #ifdef DEBUG
      printk("%s: module removed.\n", ADAPTER_ID);
  #endif

  }    
}

#ifdef __cplusplus
}
#endif

/***************************************************************************
 *
 * read() service function
 *
 ***************************************************************************/

static int dio_read(struct inode *iNode, struct file *filePtr, char *buf, int count)
{
  BYTE  value;
  int   minor = MINOR(iNode->i_rdev);

  /* check to see if reading a value from the DIO */


  if ( minor >= 0 && minor < DIO_PORTS ) {
    if ( ENABLE_INTERRUPTS && !(Chan[minor].f_flags & O_NONBLOCK)) {
          interruptible_sleep_on(&dio24_wait);           /* Block in wait state */
    }
    value = inb(Chan[minor].addr);
    put_fs_byte(value, buf);
  } else {
    printk("dio_read: Incorrect minor number (%d).\n", minor);
    return -1;
  }

  #ifdef DEBUG
  printk("dio_read: %s DIO Port %#x set to %#x\n", ADAPTER_ID, minor, value);
  #endif

  return 1;
}

/***************************************************************************
 *
 * write() service function
 *
 ***************************************************************************/

static int dio_write(struct inode *iNode, struct file *filePtr, const char *buf, int count)
{
  int  minor = MINOR(iNode->i_rdev);
  BYTE value;

  /* check to see if writing a value to the DIO */
  if ( minor >= 0 && minor <  DIO_PORTS ) {
    value = get_fs_byte(buf);
    Chan[minor].value = value;
    outb_p( value, Chan[minor].addr);
  } else {
    printk("dio_write: Incorrect minor number (%d).\n", minor);
    return -1;
  }
  #ifdef DEBUG
  printk("dio_write %s: DIO Port %#x set to %#x\n", ADAPTER_ID, minor, value);
  #endif
  return 1;
}


/***************************************************************************
 *
 * open() service handler
 *
 ***************************************************************************/

static int dio_open(struct inode *iNode, struct file *filePtr)
{
  int minor = MINOR(iNode->i_rdev);

  /* 
     check if device is already open: only one process may read from a
     port at a time.  There is still the possibility of two processes 
     reading from two different channels messing things up. However,
     the overhead to check for this may not be worth it.
  */

  if ( Chan[minor].open == TRUE ) {
      return -EBUSY;
  }

  MOD_INC_USE_COUNT;
  Chan[minor].open = TRUE;                 /* The device is open */
  Chan[minor].f_flags = filePtr->f_flags;

  #ifdef DEBUG
      printk("%s: open(): minor %d mode %d.\n", ADAPTER_ID, minor, Chan[minor].mode);
  #endif
  return 0;   
}

/***************************************************************************
 *
 * close() service handler
 *
 ***************************************************************************/

static void dio_close(struct inode *iNode, struct file *filePtr)
{
  int minor = MINOR(iNode->i_rdev);

  #ifdef DEBUG
      printk("%s: close() of minor number %d.\n", ADAPTER_ID, minor);
  #endif

  MOD_DEC_USE_COUNT;
  Chan[minor].open = FALSE;
}

/***************************************************************************
 *
 * iotctl() service handler
 *
 ***************************************************************************/

/* Note:
    Remember that FIOCLEX, FIONCLEX, FIONBIO, and FIOASYN are reserved ioctl cmd numbers
*/

static int dio_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg)
{
  int minor = MINOR(iNode->i_rdev);

  if (_IOC_TYPE(cmd) != IOCTL_MAGIC) return -EINVAL;
  if (_IOC_NR(cmd) > IOCTL_MAXNR)  return -EINVAL;

  if (minor < 0 || minor >= DIO_PORTS) {
        return(-EINVAL);
  }

  switch (cmd) {

      case DIO_SET_MODE:
        arg &= 0x3;
        switch (minor) {
	    case 0:		       /* Port A */
              #ifdef DEBUG
              printk("DIO_SET_MODE for Port A\n");
              #endif
              dio_reg &= 0x9f;
              dio_reg |= (arg << 5);
              outb_p(dio_reg, DIO_CNTRL_REG);
              Chan[minor].mode = arg;
              break;

            case 1:			/* Port 1B */
              #ifdef DEBUG
              printk("DIO_SET_MODE for Port 1B\n");
              #endif
              if (arg & 0x2)
	        return(-EINVAL);	       /* Port 1B only has Modes 0 & 1 */
              dio_reg &=  0xfb;
              dio_reg |= (arg << 2);
              outb_p(dio_reg, DIO_CNTRL_REG);
              Chan[minor].mode = arg;
              break;

            case 2:			/* Port 1C */
              #ifdef DEBUG
              printk("DIO_SET_MODE for Port 1C\n");
              #endif
              if (arg)
	        return(-EINVAL);	/* Port 1C only has Mode 0 I/O */
              Chan[minor].mode = arg;
              break;

            default:
              #ifdef DEBUG
              printk("DIO_SET_MODE for Invalid Port\n");
              #endif
              return(-EINVAL);	/* Wrong Port Number */
              break;
        }
        break;
    
      case DIO_SET_DIRECTION:

        #ifdef DEBUG
        printk("DIO_SET_DIRECTION: minor = %d, arg = %d\n", (int) minor, (int) arg );
        #endif

        switch (minor) {
          case 0:			/* Port A */
            #ifdef DEBUG
            printk("DIO_SET_DIRECTION for Port A\n");
            #endif
            arg &= 0x1;
            dio_reg &=  0xef;
            dio_reg |= (arg << 4);
            outb_p(dio_reg, DIO_CNTRL_REG);
            break;

          case 1:			/* Port B */
            #ifdef DEBUG
            printk("DIO_SET_DIRECTION for Port B\n");
            #endif
            arg &= 0x1;
            dio_reg &=  0xfd;
            dio_reg |= (arg << 1);
            outb_p(dio_reg, DIO_CNTRL_REG);
            break;

          case 2:			/* Port C */
            #ifdef DEBUG
            printk("DIO_SET_DIRECTION for Port C\n");
            #endif
            switch ( arg ) {
              case PORT_INPUT:         /* CU = IN  CL = IN */
                  dio_reg |= 0x9;
                  break;
              case PORT_OUTPUT:        /* CU = OUT  CL = OUT */
                  dio_reg &= ~(0x9);
                  break;
              case LOW_PORT_INPUT:     /* CL = IN */
                  dio_reg |= 0x1;
                  break;
              case LOW_PORT_OUTPUT:    /* CL = OUT */
                  dio_reg &= ~(0x1);
                  break;
              case HIGH_PORT_INPUT:    /* CU = IN */
                  dio_reg |= 0x8;
                  break;
              case HIGH_PORT_OUTPUT:   /* CU = OUT */
                  dio_reg &= ~(0x8);
                  break;
              default:
                  return(-EINVAL);
                  break;
            }
            outb_p(dio_reg, DIO_CNTRL_REG);
            break;

          default:
            #ifdef DEBUG
            printk("DIO_SET_DIRECTION for Invalid Port\n");
            #endif
            return(-EINVAL);
            break;
        }
        break;
    
      default:
        return(-EINVAL);
        break;
  }

  return 0;
}


/******************************
 *                            *
 *     Interrupt handler      *
 *                            *
 ******************************/

static void dio24_ReadInterrupt(int irq)
{

#ifdef DEBUG
      printk("Entering dio24_ReadInterrupt()\n");
#endif
      
  /* check to see if the PCI-DIO24 did in fact interrupt */
  if ( inl(PCI_INT_ADDR) & INT ) {
    BoardData.control_reg = inl(PCI_INT_ADDR);
    outl(BoardData.control_reg & ~INT, PCI_INT_ADDR);
    outl(BoardData.control_reg | INTE | PCIINT, PCI_INT_ADDR);
    wake_up_interruptible(&dio24_wait);
  }
}







